import traceback
import urllib.request as urllib
from collections import Counter
from datetime import datetime, timedelta, timezone
from json import loads
from pathlib import Path
from statistics import mean
from typing import Literal, Union

import pkg_resources

key_default = ""


def jsonLoad(filename: Union[str, Path]):
    filename = pkg_resources.resource_filename(__name__, str(filename))
    with open(filename, "r", encoding="utf-8") as f:
        value = loads(f.read())
        f.close()
    return value


def lc(locale: str, key: str, *args: str):
    """Get locale string

    Args:
        locale (str): Filename of locale
        key (str): String key
        *args (str): Dict path to key

    Returns:
        str: Found locale key or error
    """
    try:
        strings = (
            jsonLoad(Path("assets/icons.json"))
            if locale == "icons"
            else jsonLoad(Path(f"locales/{locale}.json"))
        )
    except:
        return f'Locale "{locale}" does not exist'
    try:
        string = strings
        for dict_key in args:
            string = string[dict_key]
        return string[key]
    except:
        return f'Key "{key}" in locale "{locale}" does not exist'


def mostCommon(lst):
    data = Counter(lst)
    return data.most_common(1)[0][0]


def between(val, minimum, maximum):
    if minimum <= val <= maximum:
        return True
    else:
        return False


def windSpeed(power, locale):
    if power <= 0.5:
        return [lc("icons", "low", "wind"), lc(locale, "calm", "wind", "strenght")]
    elif between(power, 0.5, 1.5):
        return [lc("icons", "low", "wind"), lc(locale, "light_air", "wind", "strenght")]
    elif between(power, 1.5, 3.3):
        return [
            lc("icons", "low", "wind"),
            lc(locale, "breeze_light", "wind", "strenght"),
        ]
    elif between(power, 3.3, 5.5):
        return [
            lc("icons", "low", "wind"),
            lc(locale, "breeze_gentle", "wind", "strenght"),
        ]
    elif between(power, 5.5, 7.9):
        return [
            lc("icons", "high", "wind"),
            lc(locale, "breeze_moderate", "wind", "strenght"),
        ]
    elif between(power, 8, 10.7):
        return [
            lc("icons", "high", "wind"),
            lc(locale, "breeze_fresh", "wind", "strenght"),
        ]
    elif between(power, 10.7, 13.8):
        return [
            lc("icons", "high", "wind"),
            lc(locale, "breeze_strong", "wind", "strenght"),
        ]
    elif between(power, 13.8, 17.1):
        return [
            lc("icons", "high", "wind"),
            lc(locale, "gale_near", "wind", "strenght"),
        ]
    elif between(power, 17.1, 20.7):
        return [lc("icons", "high", "wind"), lc(locale, "gale", "wind", "strenght")]
    elif between(power, 20.7, 24.4):
        return [
            lc("icons", "high", "wind"),
            lc(locale, "gale_strong", "wind", "strenght"),
        ]
    elif between(power, 24.4, 28.4):
        return [lc("icons", "high", "wind"), lc(locale, "storm", "wind", "strenght")]
    elif between(power, 28.4, 32.6):
        return [
            lc("icons", "high", "wind"),
            lc(locale, "storm_violent", "wind", "strenght"),
        ]
    elif power >= 32.6:
        return [
            lc("icons", "high", "wind"),
            lc(locale, "hurricane", "wind", "strenght"),
        ]
    else:
        return [lc("icons", "error", "wind"), lc(locale, "unknown", "wind", "strenght")]


def windDirection(degree, locale):
    if between(degree, 0, 22.4) or between(degree, 337.5, 360):
        return lc("icons", "north", "direction")
    elif between(degree, 22.5, 67.4):
        return lc("icons", "north_east", "direction")
    elif between(degree, 67.5, 112.4):
        return lc("icons", "east", "direction")
    elif between(degree, 112.5, 157.4):
        return lc("icons", "south_east", "direction")
    elif between(degree, 157.5, 202.4):
        return lc("icons", "south", "direction")
    elif between(degree, 202.5, 247.4):
        return lc("icons", "south_west", "direction")
    elif between(degree, 247.5, 292.4):
        return lc("icons", "west", "direction")
    elif between(degree, 292.5, 337.4):
        return lc("icons", "north_west", "direction")
    else:
        return lc("icons", "unknown", "direction")


def timestampModified(timestamp, offset):
    return datetime.fromtimestamp(timestamp, tz=timezone.utc) + timedelta(hours=offset)


def log_weather():
    return f"[{datetime.now().strftime('%d.%m.%Y | %H:%M:%S')}]"


# ========================================================================================================================================================
# Weather forecast for date
def newWeather(
    date,
    lat,
    lon,
    apikey=key_default,
    exchange=None,
    locale: Literal["en", "ru", "uk"] = "uk",
    test=False,
    full=True,
    separate=False,
    svodka=False,
    offset=0,
    prefix=" • ",
    suffix=";",
    connector="\n",
):
    print(f"{log_weather()} Requesting weather in {lat}, {lon} using key {apikey}")

    if svodka:
        seldate = datetime.now(tz=timezone.utc) + timedelta(hours=offset)
    else:
        seldate = datetime.now()

    url = f"https://api.openweathermap.org/data/3.0/onecall?lat={lat}&lon={lon}&exclude=minutely&units=metric&lang={locale}&appid={apikey}"

    try:
        data = loads(urllib.urlopen(url).read().decode())
    except Exception as exp:
        return {
            "weather": "",
            "gentime": seldate.strftime("%H:%M"),
            "outcome": "Failure",
            "exception": "Could not get API response",
            "traceback": str(traceback.format_exc()),
        }

    output = []

    temp_list = []
    weath_icon_list = []
    feels_list = []
    wind_list = []
    wind_direct = []
    gust_list = []
    pop_time_list = []
    pop_list = []
    rain_list = []
    humid_list = []
    uvi_list = []

    found = False

    for day in data["daily"]:
        if timestampModified(day["dt"], offset).strftime("%d.%m.%Y") == date:
            found = True

            sunrise = timestampModified(day["sunrise"], offset).strftime("%H:%M")
            sunset = timestampModified(day["sunset"], offset).strftime("%H:%M")

            break

    if not found:
        return {
            "weather": "",
            "gentime": seldate.strftime("%H:%M"),
            "umbrella": "error",
            "suncream": "error",
            "hot": "error",
            "cold": "error",
            "rainy": "error",
            "outcome": "Failure",
            "exception": "Weather for day not found",
            "traceback": "",
        }

    try:
        for hour in data["hourly"]:
            if timestampModified(hour["dt"], offset).strftime("%d.%m.%Y") == date:
                temp_list.append(round(hour["temp"]))
                weath_icon_list.append(str(int(hour["weather"][0]["icon"][:2])))
                feels_list.append(round(hour["feels_like"]))
                wind_list.append(round(hour["wind_speed"]))
                wind_direct.append(round(hour["wind_deg"], 1))
                gust_list.append(round(hour["wind_gust"]))
                pop_list.append(round(100 * (hour["pop"])))

                if int(hour["weather"][0]["icon"][:2]) in [9, 10, 11, 13]:
                    pop_time_list.append(
                        timestampModified(hour["dt"], offset).strftime("%H:%M")
                    )

                if "rain" in hour:
                    rain_list.append(hour["rain"]["1h"])

                if "snow" in hour:
                    rain_list.append(hour["snow"]["1h"])

                humid_list.append(hour["humidity"])
                uvi_list.append(round(hour["uvi"]))

        # Температура воздуха
        temp_max = str(max(temp_list))
        temp_min = str(min(temp_list))

        output.append(
            lc(locale, "exact", "temp").format(
                lc("icons", mostCommon(weath_icon_list), "sky"), temp_min, temp_max
            )
        )

        if max(temp_list) >= 32:
            hot = True
        else:
            hot = False

        if min(temp_list) < -10:
            cold = True
        else:
            cold = False

        # Температура ощущается
        feels_max = str(max(feels_list))
        feels_min = str(min(feels_list))
        # feels_icon =
        if (abs(int(feels_max) - int(temp_max)) > 2) or (
            abs(int(feels_min) - int(temp_min)) > 2
        ):
            output.append(
                lc(locale, "feels", "temp").format(
                    lc("icons", mostCommon(weath_icon_list), "sky"),
                    feels_min,
                    feels_max,
                )
            )

        # Осадки
        if max(pop_list) >= 30:
            try:
                rain_amount = max(rain_list)

                if 0.25 < rain_amount < 2.5:
                    rain_strength = lc(locale, "low", "ppt", "strenght")
                elif between(rain_amount, 2.5, 10):
                    rain_strength = lc(locale, "moderate", "ppt", "strenght")
                elif between(rain_amount, 10.01, 50):
                    rain_strength = lc(locale, "high", "ppt", "strenght")
                elif rain_amount > 50:
                    rain_strength = lc(locale, "storm", "ppt", "strenght")
                else:
                    rain_strength = ""
            except:
                rain_amount = 0
                rain_strength = lc(locale, "unknown", "ppt", "strenght")

            if between(max(pop_list), 30, 40):
                pop_icon = "☂"
            else:
                pop_icon = "☔"

            if len(pop_time_list) == 1:
                output.append(
                    lc(locale, "exact", "ppt", "chance").format(
                        pop_icon, max(pop_list), rain_strength, pop_time_list[0]
                    )
                )
            elif len(pop_time_list) > 1:
                output.append(
                    lc(locale, "timespan", "ppt", "chance").format(
                        pop_icon,
                        max(pop_list),
                        rain_strength,
                        pop_time_list[0],
                        pop_time_list[-1],
                    )
                )
            else:
                output.append(
                    lc(locale, "simple", "ppt", "chance").format(
                        pop_icon, max(pop_list), rain_strength
                    )
                )
        else:
            rain_amount = 0
            rain_strength = ""
            output.append(lc(locale, "none", "ppt", "umbrella"))  # type: ignore

        # Wind
        wind_max = max(wind_list)
        wind_min = min(wind_list)

        wind_info = windSpeed(wind_max, locale)
        wind_deg = round(mean(wind_direct), 1)
        wind_dir = windDirection(wind_deg, locale)

        if max(wind_list) == min(wind_list):
            output.append(
                lc(locale, "exact", "wind").format(
                    wind_info[0], wind_info[1], wind_max, wind_dir, int(wind_deg)
                )
            )
        else:
            output.append(
                lc(locale, "span", "wind").format(
                    wind_info[0],
                    wind_info[1],
                    wind_min,
                    wind_max,
                    wind_dir,
                    int(wind_deg),
                )
            )

        # Wind gusts
        gust_max = max(gust_list)

        if abs(gust_max - int(wind_max)) > 3:
            output.append(lc(locale, "gusts", "wind").format(gust_max))

        # Влажность
        humid = round(mean(humid_list))
        output.append(lc(locale, "humidity", "temp").format(humid))

        # УФ индекс
        uvi_max = max(uvi_list)

        if between(uvi_max, 0, 2):
            uvi_strength = lc(locale, "low", "uv", "strenght")
            uvi_icon = lc("icons", "low", "uv")
        elif between(uvi_max, 3, 5):
            uvi_strength = lc(locale, "moderate", "uv", "strenght")
            uvi_icon = lc("icons", "moderate", "uv")
        elif between(uvi_max, 6, 7):
            uvi_strength = lc(locale, "high", "uv", "strenght")
            uvi_icon = lc("icons", "high", "uv")
        elif between(uvi_max, 8, 10):
            uvi_strength = lc(locale, "very_high", "uv", "strenght")
            uvi_icon = lc("icons", "very_high", "uv")
        elif uvi_max > 10:
            uvi_strength = lc(locale, "extreme", "uv", "strenght")
            uvi_icon = lc("icons", "extreme", "uv")
        else:
            uvi_strength = lc(locale, "unknown", "uv", "strenght")
            uvi_icon = lc("icons", "unknown", "uv")

        output.append(
            lc(locale, "status", "uv").format(uvi_icon, uvi_strength, uvi_max)
        )

        # Рассвет и закат
        output.append(lc(locale, "sun").format(sunrise, sunset))  # type: ignore

        # Курс доллара
        if full and exchange != None:
            output.append(exchange("Доллар", "usd", "💵", test=test))
            output.append(exchange("Евро", "eur", "💶", test=test))

        # Зонтик
        if between(max(pop_list), 30, 40) and rain_amount >= 1:
            umbrella = lc(locale, "low", "ppt", "umbrella")
        elif between(max(pop_list), 41, 60) and rain_amount >= 1:
            umbrella = lc(locale, "moderate", "ppt", "umbrella")
        elif between(max(pop_list), 61, 80) and rain_amount >= 1:
            umbrella = lc(locale, "high", "ppt", "umbrella")
        elif max(pop_list) > 80 and rain_amount >= 1:
            umbrella = lc(locale, "sure", "ppt", "umbrella")
        else:
            umbrella = ""

        if max(pop_list) > 80 and rain_amount >= 1:
            rainy = True
        else:
            rainy = False

        # Защита от солнца
        if uvi_max == 4:
            suncream = lc(locale, "moderate", "uv", "suncream")
        elif uvi_max == 5:
            suncream = lc(locale, "high", "uv", "suncream")
        elif 5 < uvi_max < 8:
            suncream = lc(locale, "very_high", "uv", "suncream")
        elif 7 < uvi_max:
            suncream = lc(locale, "extreme", "uv", "suncream")
        else:
            suncream = ""

        return {
            "weather": connector.join([prefix + entry + suffix for entry in output]),
            "gentime": seldate.strftime("%H:%M"),
            "umbrella": umbrella,
            "suncream": suncream,
            "hot": hot,
            "cold": cold,
            "rainy": rainy,
            "outcome": "Success",
            "exception": "",
            "traceback": "",
        }

    except Exception as exp:
        traceback.print_exc()
        return {
            "weather": prefix + lc(locale, "exception", "app").format(exp) + suffix,
            "gentime": seldate.strftime("%H:%M"),
            "umbrella": "error",
            "suncream": "error",
            "hot": "error",
            "cold": "error",
            "rainy": "error",
            "outcome": "Failure",
            "exception": exp,
            "traceback": str(traceback.format_exc()),
        }
